import java.util.Arrays;

public class Simplex {
    private final boolean debugging; // booléen de debug
    private final Matrix matEcart; // matrice d'écart
    private final Matrix tabAux; // matrice auxiliaire (tableau auxiliaire en phase 1)
    private final int nbSousCondition;
    private final int ligne; // Ligne
    private final int colonne; // Colonne
    private int nbPivot;
    private static final double EPSILON = 0E-7;
    private final int nbContraintes;

    public Matrix getMatEcart() {
        return matEcart;
    }

    public int getNbPivot() {
        return nbPivot;
    }

    /**
     * Constructeur du simplex
     * @param ligne le nombre de lignes dans le tableau initial
     * @param colonne le nombre de colonnes dans le tableau initial
     * @param nbSousCondition le nombre de sous condition
     * @param nbContraintes le nombre de contraintes
     * @param debugging mode de debug
     */
    public Simplex(int ligne, int colonne, int nbSousCondition, int nbContraintes, boolean debugging) {
        this.ligne = ligne;
        this.colonne = colonne;
        this.matEcart = new Matrix(ligne, colonne);
        this.nbSousCondition = nbSousCondition;
        this.tabAux = new Matrix(ligne + 2, colonne + nbSousCondition + 1);
        this.nbPivot = 0;
        this.nbContraintes = nbContraintes;
        this.debugging = debugging;
    }

    /**
     * Crée la matrice d'écart pour le début de la résolution du simplex (transformation du tableau initial en matrice)
     * @param eq Equation retournée par le parser
     * @param nbContraintes nombre de contraintes
     */
    void createSimplex(Equation eq, int nbContraintes) {
        // Matrice Amxn
        for (int i = 0; i < this.ligne; i++) {
            for (int j = 0; j < this.colonne; j++) {
                if (j < nbContraintes) {
                    this.matEcart.setData(i, j, eq.getMatAtId(i, j));
                } else {
                    this.matEcart.setData(i, j, j - nbContraintes == i ? 1.0 : 0.0);
                }
            }
        }

        // Ajout des membres de droites
        for (int i = 0; i <= this.ligne - 1; i++)
        {
            this.matEcart.setData(i, this.colonne - 1, eq.getRightVec().get(i));
        }


        // Ajout de la fonction objective
        this.matEcart.matrixRealloc(this.matEcart.getLine() + 1, this.matEcart.getCol());
        for (int i = 0; i < nbContraintes; i++)
        {
            this.matEcart.setData(this.ligne, i, eq.getFuncObj().get(i));
        }
    }

    /**
     * Choix de la phase
     * Si b[i] < 0 phase 1 sinon phase 2
     *
     * @return true = phase 1 | false = phase 2
     */
    boolean which_phase() {
        for (int i = 0; i < this.ligne; i++) {
            if (signe(this.matEcart.getData(i, this.colonne - 1))) return true;
        }
        return false;
    }

    /**
     * Phase 1
     * @return le résultat du pivot
     */
    Matrix tabAux() {
        if (debugging) System.out.println();
        double[] tabRes = new double[this.colonne + this.nbSousCondition + 1];
        Arrays.fill(tabRes, 1.0);
        // Regarde si un membre de droite est négatif, si c'est le cas, on inverse toute la ligne
        for (int i = 0; i < this.colonne; i++) {
            for (int j = 0; j < this.ligne; j++) {
                if (this.matEcart.getData(j, this.colonne - 1) < 0) {
                    if (this.matEcart.getData(j, i) != 0) {
                        this.matEcart.setData(j, i, -this.matEcart.getData(j, i));
                    }
                }
            }
        }
        // Calcul des coûts
        for (int j = 0; j < this.colonne; j++) {
            double tmpSum = 0;
            for (int i = 0; i < this.ligne; i++) {
                tmpSum += -1 * this.matEcart.getData(i, j);
            }
            tabRes[j] = tmpSum;
        }

        // -2 car => tabRes[(tabRes.length-1) - (this.nbSousCondition - 1)];
        // Création du tableau auxiliaire
        tabRes[tabRes.length - 1] = tabRes[tabRes.length - this.nbSousCondition - 2];
        tabRes[tabRes.length - this.nbSousCondition - 2] = 0;
        for (int i = 0; i < this.tabAux.getLine(); i++) {
            for (int j = 0; j < this.tabAux.getCol(); j++) {
                if (i < this.matEcart.getLine() && j < this.matEcart.getCol()) {
                    this.tabAux.setData(i, j, this.matEcart.getData(i, j));
                } else if (i == this.tabAux.getLine() - 1) {
                    this.tabAux.setData(i, j, tabRes[j]);
                } else // membre de droite à la fin du tab
                {
                    this.tabAux.setData(i, j, this.matEcart.getData(i, j - this.matEcart.getLine()));
                }
            }
        }
        if (debugging) this.tabAux.printTabAux("Tableau auxiliaire", this.nbContraintes, this.nbSousCondition,
                this.tabAux.getCol() - this.matEcart.getCol() - 1);

        // Pivotage sur le tableau auxiliaire
        pivot(this.tabAux, true);
        double solutionOptimale = this.tabAux.getData(this.tabAux.getLine() - 1, this.tabAux.getCol() - 1);
        // Si la solution optimale est positive, pas de solutions
        if (solutionOptimale > EPSILON) {
            System.out.println("Il n'y a pas de solutions admissibles pour ce problème.");
        }
        // si la solution optimale est ~= 0, il faut pivoter sur le "petit tableau"
        if (Math.abs(solutionOptimale) < EPSILON || solutionOptimale == 0) {
            // Il y a une solution optimale.
            // Il faut enlever les variables auxilaires
            Matrix res = new Matrix(matEcart.getLine(), matEcart.getCol()); // pour récupérer la taille de la matrice d'écart
            // remplir la nouvelle matrice avec les valeurs de la matrice auxiliaire, mais dans la taille de la matrice d'écart
            res.matrixFill(res.getLine(), res.getCol(), tabAux.getDatas());
            if (debugging) res.matrixPrint("Petit tableau");
            nbPivot = 0; // reset du nombre de pivots, on pourrait implémenter un pivot de phase 1 et un pivot de phase 2
            // retour a la phase 2 avec le pivot sur la nouvelle matrice
            pivot(res, true);
            if (debugging) {
                res.matrixPrint("Matrice Résultat ");
                System.out.println("Nombre de pivot : " + nbPivot);
            }
            return res;
        }
        return null;
    }

    /**
     * Sélectionne le premier négatif rencontré sur la ligne de la fonction objective
     * @param mat la matrice
     * @return la colonne du premier négatif rencontré
     */
    int getFirstNeg(Matrix mat) {
        for (int j = 0; j < mat.getCol() - 1; j++) {
            if (signe(mat.getData(mat.getLine() - 1, j))) return j;
        }
        return -1;
    }

    /**
     * Fonction de pivot
     * @param mat phase 1 ⇒ tableau auxiliaire | phase 2 => Matrice d'écart
     * @param phase true => phase 1 | false => phase 2
     */
     void pivot(Matrix mat, boolean phase) {
        this.nbPivot += 1;
        // Ligne du négatif
        int firstNeg = getFirstNeg(mat);
        if (firstNeg == -1) return; // si pas de négatif
        boolean has_neg = false;
        // Selection de la colonne du pivot
        int id = ligneSortante(mat, firstNeg, phase);
        // Selection de la valeur du pivot
        double val_pivot = mat.getData(id, firstNeg);

        // Application du pivot de gauss sur la ligne du pivot, pour avoir une "référence" de coefficient pour la suite
        for (int i = 0; i < mat.getCol(); i++) {
            mat.setData(id, i, mat.getData(id, i) / val_pivot);
        }

        // Application du pivot de gauss sur toute la matrice (sauf la ligne du pivot précédemment appliquée)
        for (int i = 0; i < mat.getLine(); i++) {
            val_pivot = mat.getData(i, firstNeg);
            for (int j = 0; j < mat.getCol(); j++) {
                if (i != id) {
                    mat.setData(i, j, mat.getData(i, j) - val_pivot * mat.getData(id, j));
                }
            }
        }
        if (debugging) {
            mat.matrixPrint("Pivot numéro " + this.nbPivot);
            System.out.println("colonne du pivot: " + firstNeg);
            System.out.println("ligne du pivot: " + id);
            System.out.println("Valeur du pivot: " + val_pivot);
        }
        // Vérification des variables en bases, si un négatif est trouvé, on peut relancer la phase 2, les opérations sont terminées
        for (int j = 0; j < mat.getCol(); j++)
            if (signe(mat.getData(mat.getLine() - 1, j))) {
                has_neg = true;
                break;
            }
        if (has_neg)
            pivot(mat, phase);
    }

    /**
     * Récupérer l'id de la ligne du pivot
     * @param mat le tableau auxiliaire
     * @param y la colonne où l'on va chercher le pivot
     * @param phase la phase dans laquelle on se situe
     * @return l'id de la ligne du pivot
     */
    int ligneSortante(Matrix mat, int y, boolean phase) {
        // phase 1 on ne va pas chercher dans les deux dernières lignes
        // phase 2, on ne va pas chercher dans la dernière ligne
        int depth = phase ? mat.getLine() - 2 : mat.getLine() - 1;
        // On commence à chercher à l'id 0
        int id = 0;
        // La première vérification permet de s'assurer qu'on ne tombe pas sur un 0
        // La 2ᵉ vérification permet de vérifier qu'au plus profond de la matrice il s'agit bien d'un négatif
        while (!(mat.getData(id, y) > 0) && mat.getData(id, depth) >= 0) {
            id++;
        }

        // On divise le membre de droite par la valeur de la matrice en question
        double tmp = mat.getData(id, mat.getCol() - 1) / mat.getData(id, y);
        double tmp_s;
        // On parcourt toutes la ligne
        for (int i = 1; i < depth; i++) {
            // Si la valeur est plus grande que 0
            if (mat.getData(i, y) > 0) {
                // On divise le membre de droite par la valeur de la matrice en question
                tmp_s = mat.getData(i, mat.getCol() - 1) / mat.getData(i, y);
                // Si la valeur actuelle est plus petite que la valeur stockée précedement,
                // on sauvegarde l'id de celui-ci
                if (tmp_s < tmp) {
                    id = i;
                    tmp = tmp_s;
                }
            }
        }
        // On retourne l'id
        return id;
    }

    /**
     * Print les informations des matrices pour le mode débogage
     * @param mat la matrice souhaitée
     * @param s le titre
     */
    public void printSimplex(Matrix mat, String s) {
        mat.matrixPrint(s);
    }

    /**
     * Print du résultat comme dans l'énnoncé
     * @param m la matrice
     * @param time le temps de l'exécution
     */
    public void resultPrint(Matrix m, long time) {
        int line = m.getLine() - 1;
        int col = m.getCol() - 1;
        System.out.println("-------------------Solution-------------------");
        System.out.println("Status: Optimal");
        System.out.format("Size: %d rows, %d cols\n", m.getLine(), m.getCol());
        System.out.println("Obj value: " + -1 * m.getData(line, col));
        System.out.println("Nombre de pivot(s) réalisé(s) sur le petit tableau post phase 1: " + this.nbPivot);
        System.out.println("Time: " + time + " ms");
        System.out.println("-------------------Variables------------------");
        System.out.println("Variables:");
        for (int i = 0; i < line; i++) {
            System.out.format("X[%d]:               =           %.4f        \n", i, m.getData(i, col));
        }

        for (int i = 0; i < col; i++) {
            System.out.format("Z_Cstr_%d:           =           %.4f        \n", i, m.getData(line, i));
        }
    }

    boolean signe(Double x) {
        return x < 0;
    }
}
